﻿using Myriadbits.NTFS;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Myriadbits.DiskSpaceAnalyzer
{

	/// <summary>
	/// Simple class that represents a single file
	/// </summary>
	public class FileData
	{
		private string m_fullPath = "";
		private string m_iconType = "";
		private int m_iconIndex = -1;
		private NTFSAttributeStandardInformation m_standardInformation = null;

		/// <summary>
		/// Constructor, create a new file data object
		/// </summary>
		/// <param name="folder"></param>
		/// <param name="index"></param>
		/// <param name="volume"></param>
		public FileData(Folder folder, int index)
		{
			this.Folder = folder;
			this.m_iconType = string.Empty;
			this.Index = index;
			NTFSVolume currentVolume = folder.Parser.Volume;

			if (index < this.Folder.ChildCount)
			{
				// It is a folder
				this.IsFolder = true;

				Folder child = this.Folder.Children[index];				
				this.ReferenceNumber = child.ReferenceNumber;
				this.FileReferenceId = this.ReferenceNumber & Myriadbits.NTFS.NTFS.referenceMask;
				this.FileName = child.Name;
				this.FileSize = child.Size;
				this.FileCount = child.NumberOfFiles;
				this.m_fullPath = child.FullPath;

				// Get the icon type
				if (Directory.Exists(this.FullPath))
				{
					this.m_iconType = FileIcon.GetIconType(this.FullPath);
				}

				
			}
			else
			{
				// It is a file
				this.IsFolder = false;

				ulong fileIndex = (ulong)(index - this.Folder.ChildCount);
				ulong fileReference = this.Folder.Files.Keys.ToArray()[fileIndex];

				this.File = this.Folder.Files[fileReference];
				ulong referenceNumberCode = this.File.ReferenceNumber;
				this.FileSize = this.File.Size;
				this.FileName = this.File.Name;

				this.ReferenceNumber = referenceNumberCode;
				this.FileReferenceId = referenceNumberCode & Myriadbits.NTFS.NTFS.referenceMask;
				this.NameIndex = (byte)((referenceNumberCode & Myriadbits.NTFS.NTFS.indexMask) >> 48);

				//NTFSFileRecord filerecord = currentVolume.ReadFileRecord(this.FileReferenceId);
				//if (filerecord != null)
				//{
				//	//this.FileName = filerecord.GetFilename(this.NameIndex);
				//	this.m_standardInformation = filerecord.GetStandardInformation(currentVolume);
				//}

				// Get the icon type
				if (System.IO.File.Exists(this.FullPath))
				{
					this.m_iconType = FileIcon.GetIconType(this.FullPath);
					if (this.m_iconType == "Application")
					{
						this.m_iconType = string.Format("{0}-{1}", this.m_iconType, this.FileName);
					}
				}
			}
		}


		/// <summary>
		/// Constructor, create a new file data object for a folder
		/// </summary>
		/// <param name="folder"></param>
		/// <param name="index"></param>
		/// <param name="volume"></param>
		public FileData(Folder folder, Folder parentFolder)
		{
			this.Folder = parentFolder;
			this.m_iconType = string.Empty;
			this.Index = -1;
			NTFSVolume currentVolume = folder.Parser.Volume;

			// It is a folder
			this.IsFolder = true;

			this.ThisFolder = folder;
			this.ReferenceNumber = this.ThisFolder.ReferenceNumber;
			this.FileReferenceId = this.ReferenceNumber & Myriadbits.NTFS.NTFS.referenceMask;
			this.FileName = this.ThisFolder.Name;
			this.FileSize = this.ThisFolder.Size;
			this.FileCount = this.ThisFolder.NumberOfFiles;
			this.m_fullPath = this.ThisFolder.FullPath;
		}
		

		/// <summary>
		/// Constructor, create a new file data object for a file
		/// </summary>
		/// <param name="folder"></param>
		/// <param name="index"></param>
		/// <param name="volume"></param>
		public FileData(Myriadbits.NTFS.File file, Folder parentFolder)
		{
			this.Folder = parentFolder;
			this.m_iconType = string.Empty;
			this.Index = -1;
			NTFSVolume currentVolume = this.Folder.Parser.Volume;
		
			// It is a file
			this.IsFolder = false;

			this.File = file;
			ulong referenceNumberCode = this.File.ReferenceNumber;
			this.FileSize = this.File.Size;
			this.FileName = this.File.Name;

			this.ReferenceNumber = referenceNumberCode;
			this.FileReferenceId = referenceNumberCode & Myriadbits.NTFS.NTFS.referenceMask;
			this.NameIndex = (byte)((referenceNumberCode & Myriadbits.NTFS.NTFS.indexMask) >> 48);
		}


		/// <summary>
		/// Load the icon
		/// </summary>
		public void LoadIcon()
		{
			if (this.IsFolder)
			{
				// Get the icon type
				//if (Directory.Exists(this.FullPath))
				{
					this.m_iconType = FileIcon.GetIconType(this.FullPath);
				}
			}
			else
			{
				// Get the icon type
				//if (System.IO.File.Exists(this.FullPath))
				{
					this.m_iconType = FileIcon.GetIconType(this.FullPath);
					if (this.m_iconType == "Application")
					{
						this.m_iconType = string.Format("{0}-{1}", this.m_iconType, this.FileName);
					}
				}
			}
		}


		/// <summary>
		/// Return the full path to this file
		/// </summary>
		public string FullPath
		{
			get 
			{
				if (string.IsNullOrEmpty(this.m_fullPath))
				{
					try
					{
						this.m_fullPath = Path.Combine(this.Folder.FullPath, this.FileName);
					}
					catch (Exception ex)
					{
						Trace.WriteLine(string.Format("Error in Fullpath: {0} --> File[{1}]: {2}", ex.Message, this.FileReferenceId, this.FileName));
					}
				}
				return this.m_fullPath;
			}
		}			


		/// <summary>
		/// Add this icon to the image list
		/// - Start by first checking if the icon is already in the image list, if it is, return that index,
		/// - Not in imagelist, add to the image list and return its index
		/// </summary>
		/// <param name="images"></param>
		public void AddIconToImageList(ImageList images)
		{
			if (!string.IsNullOrEmpty(this.m_iconType))
			{
				if (!images.Images.ContainsKey(this.m_iconType))
				{
					// Load the icon type
					Icon iconNew = FileIcon.GetSmallIcon(this.FullPath);					
					if (iconNew != null)
						images.Images.Add(this.m_iconType, iconNew);
				}
			}
		}

		/// <summary>
		/// Return (store internally) the icon index
		/// </summary>
		/// <param name="images"></param>
		/// <returns></returns>
		public int GetImageIndex(ImageList images)
		{
			if (m_iconIndex == -1)
				m_iconIndex = images.Images.IndexOfKey(this.IconType);
			return m_iconIndex;
		}

		public string IconType { get { return m_iconType; } }
		public int Index { get; set; }
		public Folder Folder { get; set; }
		public ulong FileReferenceId { get; set; }
		public ulong ReferenceNumber { get; set; }
		public string FileName { get; set; }
		public ulong FileSize { get; set; }
		public int FileCount { get; set; }
		public byte NameIndex { get; set; }
		public bool IsFolder { get; set; }
		public Myriadbits.NTFS.File File { get; set; }
		public Folder ThisFolder { get; set; }

		/// <summary>
		/// Wrapper around filerank, unkown rank for folders and unknown ranking
		/// </summary>
		public int FileRank
		{
			get
			{
				if (this.File != null && this.File.Rank != -1)
					return this.File.Rank;
				return Myriadbits.NTFS.File.UnknownRank; // Extreme high => better sorting results
			}
		}

		/// <summary>
		/// Returns true if there is a standardinformation block present for this file (almost all files have one!)
		/// CALLING THIS METHOD WILL LOAD THE STANDARDINFORMATION FROM DISK (=SLOW)
		/// </summary>
		public bool HasStandardInformation
		{
			get
			{
				NTFSAttributeStandardInformation standardInformation = StandardInformation;
				return (standardInformation != null);
			}
		}


		/// <summary>
		/// Returns the last modification time
		/// CALLING THIS METHOD WILL LOAD THE STANDARDINFORMATION FROM DISK (=SLOW)
		/// </summary>
		public DateTime LastModified
		{
			get
			{
				NTFSAttributeStandardInformation standardInformation = this.StandardInformation;
				if (standardInformation != null)
					return this.StandardInformation.FileAlteredTime;
				return new DateTime(0);
			}
		}

		/// <summary>
		/// Return the standardinformation, when not yet loaded, load it from the NTFS file system
		/// CALLING THIS METHOD WILL LOAD THE STANDARDINFORMATION FROM DISK (=SLOW)
		/// </summary>
		public NTFSAttributeStandardInformation StandardInformation 
		{ 
			get 
			{
				if (this.m_standardInformation == null)
				{
					NTFSVolume currentVolume = this.Folder.Parser.Volume;
					NTFSFileRecord filerecord = currentVolume.ReadFileRecord(this.FileReferenceId);
					if (filerecord != null)
					{
						this.m_standardInformation = filerecord.GetStandardInformation(currentVolume);
					}
				}
				return this.m_standardInformation; 
			} 
		}


	}
}
