﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Myriadbits.NTFS
{
	/// <summary>
	/// Main class that represents a single directory/folder
	/// Every folder can have multiple child folders
	/// and may in itself contain files.
	/// </summary>
	public class Folder
	{
		public NTFSReader Parser { get; set; }

		public ulong ReferenceNumber { get; set; }
		public Folder ParentFolder { get; set; }
		public string Name { get; set; }
		public ulong Size { get; set; }
		public int NumberOfFiles { get; set; }
		public int OwnNumberOfFiles { get; set; }
		public ulong OwnSize { get; set; }

		public List<Folder> Children { get; set; }
		public Dictionary<ulong, File> Files { get; set; }

		public Folder()
		{
			this.Children = null;
			this.Files = null;
			this.ParentFolder = null;
			this.ReferenceNumber = 0;
			this.Size = 0;
			this.NumberOfFiles = 0;
			this.OwnSize = 0;
			this.OwnNumberOfFiles = 0;
		}

		public Folder(ulong refNumber, ulong dataSize = 0, int numberOfFiles = 0)			
		{
			this.Children = null;
			this.Files = null;
			this.ParentFolder = null;
			this.ReferenceNumber = refNumber;
			this.Size = dataSize;
			this.NumberOfFiles = numberOfFiles;
			this.OwnSize = 0;
			this.OwnNumberOfFiles = 0;
		}


		public Folder(ulong refNumber, NTFSAttributeFilename nameAttr, NTFSAttributeStandardInformation standardAttr)
		{
			this.Children = null;
			this.Files = null;
			this.ParentFolder = null;
			this.ReferenceNumber = refNumber;
			if (nameAttr != null)
				this.Name = nameAttr.Filename;
			this.Size = 0;
			this.NumberOfFiles = 0;
			this.OwnSize = 0;
			this.OwnNumberOfFiles = 0;
			//if (standardAttr != null)
			//	this.DOSFilePermissions = standardAttr.DOSFilePermissions;
			//else
			//	this.DOSFilePermissions = 0;
		}


		/// <summary>
		/// Return the full path of a folder by traversing its parents
		/// </summary>
		public string FullPath
		{
			get 
			{
				string path = ""; // Start out empty
				Folder currentFolder = this;
				while (currentFolder != null)
				{
					if (currentFolder.ParentFolder == null)
						path = string.Format("{0}:\\{1}", currentFolder.Name, path);
					else
						path = string.Format("{0}\\{1}", currentFolder.Name, path);
					currentFolder = currentFolder.ParentFolder;
				}
				return path;
			}
		}


		/// <summary>
		/// Add a child to a parent
		/// </summary>
		/// <param name="child"></param>
		public void AddChild(Folder child)
		{
			if (this.Children == null)
				this.Children = new List<Folder>();
			this.Children.Add(child); // Add
		}

		/// <summary>
		/// Remove a child from a parent
		/// </summary>
		/// <param name="child"></param>
		public void RemoveChild(Folder child)
		{
			if (this.Children != null)
				this.Children.Remove(child);
		}

		/// <summary>
		/// Returns true if a folder contains subfolders
		/// </summary>
		public bool HasChildren 
		{ 
			get
			{
				return (this.Children != null && this.Children.Count > 0);
			}
		}

		/// <summary>
		/// Returns the number of children
		/// </summary>
		public int ChildCount
		{
			get
			{
				if (this.Children != null)
					return this.Children.Count;
				return 0;
			}
		}


		/// <summary>
		/// Returns the number of files
		/// </summary>
		public int FileCount
		{
			get
			{
				if (this.Files != null)
					return this.Files.Count;
				return 0;
			}
		}


		/// <summary>
		/// Add a file to this folder
		/// Note that if the file already exists, it will NOT be added
		/// (This method should be as fast as possible!)
		/// </summary>
		/// <param name="referenceNumber"></param>
		/// <param name="fileSize"></param>
		public void AddFile(ulong referenceNumber, ulong fileSize, string fileName)
		{
			if (this.Files == null)
				this.Files = new Dictionary<ulong, File>();

			if (this.Files.ContainsKey(referenceNumber))
			{
				// Already exists, do NOT add
				return;
			}
			
			// Update counters
			this.Size += fileSize;
			this.NumberOfFiles++;
			this.OwnSize += fileSize;
			this.OwnNumberOfFiles++;

			// Add the file
			this.Files.Add(referenceNumber, new File(this, referenceNumber, fileSize, fileName));

			if (this.ParentFolder != null)
			{
				// Update the parent's (and grandparents) sizes as well
				this.ParentFolder.IncreaseData(fileSize, 1);
			}
		}


		/// <summary>
		/// Remove a file from the list (use after deleting a file)
		/// </summary>
		/// <param name="referenceNumber"></param>
		/// <param name="fileSize"></param>
		public bool RemoveFile(ulong referenceNumber)
		{
			if (this.Files == null)
				return false; // File does not exist in this folder

			if (!this.Files.ContainsKey(referenceNumber))
				return false; // File cannot be foung

			File fd = this.Files[referenceNumber];
			ulong fileSize = 0;
			if (fd != null)
			{
				fileSize = fd.Size;
				this.Files.Remove(referenceNumber);
			}

			// Decrease the filesizes of our parents
			this.DecreaseData(fileSize, 1);

			if (this.Files.Count == 0)
			{
				this.Files = null;
			}
			return true;
		}


		/// <summary>
		/// Returns true if this folder has files
		/// </summary>
		public bool HasFiles
		{
			get
			{
				if (this.Files != null && this.Files.Count > 0)
					return true;
				return false;
			}
		}


		/// <summary>
		/// Create a string representation of this folder
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return string.Format("{0}: {1}", this.ReferenceNumber.ToString(), this.Name);
		}


		/// <summary>
		/// Set the new parent
		/// </summary>
		/// <param name="newParent"></param>
		public void SetParent(Folder newParent)
		{
			if (newParent == null)
				return;

			if (newParent == this)
				return; 

			if (this.ParentFolder != null && this.ParentFolder != newParent)
			{
				// Hmm, strange, different parent, sub from old, add to new
				// TODO Don't know if this ever happens, check
				this.ParentFolder.DecreaseData(this.Size, this.NumberOfFiles);

				// Remove ourselfs from the old parent
				this.ParentFolder.RemoveChild(this);
			}

			// If we didn't have a parent, set the new parent
			if (this.ParentFolder == null)
			{
				// Previously unknown
				this.ParentFolder = newParent;
				this.ParentFolder.IncreaseData(this.Size, this.NumberOfFiles);

				// We are now a child of the parent
				this.ParentFolder.AddChild(this);
			}
		}


		/// <summary>
		/// Increase the data of this folder and all (known) parents
		/// </summary>
		/// <param name="dataSize"></param>
		/// <param name="numberOfFiles"></param>
		public void IncreaseData(ulong dataSize, int numberOfFiles)
		{
			this.Size += dataSize;
			this.NumberOfFiles += numberOfFiles;
			
			// Call the parent recursively
			if (this.ParentFolder != null)
			{			
				this.ParentFolder.IncreaseData(dataSize, numberOfFiles);
			}
		}


		/// <summary>
		/// Decrease the data of this folder and all (known) parents
		/// </summary>
		/// <param name="dataSize"></param>
		/// <param name="numberOfFiles"></param>
		public void DecreaseData(ulong dataSize, int numberOfFiles)
		{
			this.Size -= dataSize;
			this.NumberOfFiles -= numberOfFiles;

			// Call the parent recursively
			if (this.ParentFolder != null)
			{
				this.ParentFolder.DecreaseData(dataSize, numberOfFiles);
			}
		}


		/// <summary>
		/// Returns the percentage as fraction of the total disk
		/// </summary>
		public double TotalSizePercentage
		{ 
			get
			{
				if (this.Parser != null)
				{
					double totalSize = (double) this.Parser.RootFolder.Size;
					if (totalSize == 0)	totalSize = 1;
					return ((double) this.Size) / totalSize;
				}
				return 0;
			}
		}


		/// <summary>
		/// Returns the percentage as fraction of the total disk using a log
		/// </summary>
		public double TotalSizePercentageLog(ulong maxRange)
		{
			if (this.Parser != null)
			{
				if (this.Size == 0) return 0;
				double divider = 1024;
				double totalSize = Math.Log10(((double)maxRange) / divider);
				if (totalSize == 0) totalSize = 1;
				return Math.Log10(((double)this.Size) / divider) / totalSize;
			}
			return 0;
		}

		
		/// <summary>
		/// Returns the percentage of owned files as fraction of the total disk
		/// </summary>
		public double TotalOwnSizePercentage
		{ 
			get
			{
				if (this.Parser != null)
				{
					double totalSize = (double) this.Parser.RootFolder.Size;
					if (totalSize == 0)	totalSize = 1;
					return ((double) this.OwnSize) / totalSize;
				}
				return 0;
			}
		}

		/// <summary>
		/// Returns the percentage of owned files as fraction of the total disk
		/// </summary>
		public double TotalOwnSizePercentageLog(ulong maxRange)
		{
			if (this.Parser != null)
			{
				if (this.OwnSize == 0) return 0;
				double divider = 1024;
				double totalSize = Math.Log10(((double)maxRange) / divider);
				if (totalSize == 0) totalSize = 1;
				return Math.Log10(((double)this.OwnSize) / divider) / totalSize;
			}
			return 0;
		}

		/// <summary>
		/// Returns the percentage as fraction of the parent size
		/// </summary>
		public double ParentSizePercentage
		{
			get
			{
				if (this.ParentFolder != null)
				{
					double totalSize = (double)this.ParentFolder.Size;
					if (totalSize == 0) totalSize = 1;
					return ((double)this.Size) / totalSize;
				}
				return 1; // Since this is the root folder show full bar
			}
		}


		/// <summary>
		/// Returns the percentage of owned files as fraction of the total disk
		/// </summary>
		public double ParentOwnSizePercentage
		{
			get
			{
				if (this.ParentFolder != null)
				{
					double totalSize = (double)this.ParentFolder.Size;
					if (totalSize == 0) totalSize = 1;
					return ((double)this.OwnSize) / totalSize;
				}
				return 0;
			}
		}

		/// <summary>
		/// Returns true if at least one of the files in this folder is larger then the minimum size
		/// OR IF one of the childs has a file of at least that size
		/// </summary>
		/// <param name="minimumSize"></param>
		/// <returns></returns>
		public bool HasFileLargerThenSize(ulong minimumSize)
		{
			if (this.Files != null)
				foreach(KeyValuePair<ulong, File> fd in this.Files)
					if (fd.Value.Size >= minimumSize)
						return true;

			// Nothing found in this folder, check the childs:
			if (this.Children != null)
			{
				foreach (Folder child in this.Children)
				{
					if (child.HasFileLargerThenSize(minimumSize))
						return true;
				}
			}

			// Still nothing found
			return false;
		}


		/// <summary>
		/// Returns true if at least one of the files in this folder is in the list of largest files
		/// OR IF one of the childs has one of those files
		/// </summary>
		/// <param name="minimumSize"></param>
		/// <returns></returns>
		public bool HasFileFromCollection(List<ulong> largestFiles)
		{
			if (this.Files != null)
				if (this.Files.Keys.Intersect(largestFiles).Count() > 0)
					return true;

			// Nothing found in this folder, check the childs:
			if (this.Children != null)
			{
				foreach (Folder child in this.Children)
				{
					if (child.HasFileFromCollection(largestFiles))
						return true;
				}
			}

			// Still nothing found
			return false;
		}
	}
}
