﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Myriadbits.NTFS
{
	/// <summary>
	/// Holds a collection of folder objects
	/// </summary>
	public class FolderCollection
	{
		private Dictionary<ulong, Folder> allFolders = null;
		private Folder rootFolder = null;


		/// <summary>
		/// Constructor, create the folder dictionary
		/// </summary>
		public FolderCollection()
		{
			this.allFolders = new Dictionary<ulong, Folder>();
		}


		/// <summary>
		/// Returns the rootfolder (or null)
		/// </summary>
		public Folder RootFolder { get { return this.rootFolder; } }

		/// <summary>
		/// Returns the total size of all files
		/// </summary>
		public ulong TotalSize
		{
			get
			{
				if (this.rootFolder != null)
					return this.rootFolder.Size;
				return 0;
			}
		}

		/// <summary>
		/// Returns the total number of files
		/// </summary>
		public int TotalFileCount
		{
			get
			{
				if (this.rootFolder != null)
					return this.rootFolder.NumberOfFiles;
				return 0;
			}
		}

		/// <summary>
		/// Returns the total number of folders
		/// </summary>
		public int TotalFolderCount
		{
			get
			{
				return this.allFolders.Count;
			}
		}

		/// <summary>
		/// Returns a specific folder (or null if not found)
		/// </summary>
		/// <param name="refNumber"></param>
		/// <returns></returns>
		public Folder this[ulong refNumber]
		{
			get
			{
				if (refNumber == 0)
					return null;
				if (this.allFolders.ContainsKey(refNumber))
					return this.allFolders[refNumber];
				return null;				
			}
		}


		/// <summary>
		/// Add a folder to this collection
		/// </summary>
		/// <param name="filerecord"></param>
		/// <param name="folderPermissionMask"></param>
		/// <returns></returns>
		public bool AddFolder(NTFSFileRecord filerecord, NTFSAttributeStandardInformation standardAttr, int folderPermissionMask)
		{
			//
			// Directory
			//
			// Add folders that are hidden!
			//if (standardAttr != null)
			{
				ulong referenceNumber = filerecord.ReferenceNumber;

				NTFSAttributeFilename nameAttr = filerecord.NameAttribute;
				if (nameAttr != null && referenceNumber != 11)
				{
					//if ((standardAttr.DOSFilePermissions & folderPermissionMask) == 0 || 
					//if (nameAttr.Filename == ".")
					{
						// If we know the new parent...
						Folder parentFolder = this[nameAttr.ParentReference];
						if (parentFolder == null)
						{
							// Parent folder is not (yet) known, create it
							// The parentfolder will later (hopefully) be overwritten with real valid data
							parentFolder = new Folder(nameAttr.ParentReference);
							this.allFolders.Add(nameAttr.ParentReference, parentFolder);
						}

						Folder currentFolder = this[referenceNumber];
						if (currentFolder != null)
						{
							// Already exists, update info
							currentFolder.Name = nameAttr.Filename;
							currentFolder.SetParent(parentFolder); // Parent might have changed, set it here
							return false;
						}
						else
						{
							// Folder does not exist, create it
							currentFolder = new Folder(referenceNumber, nameAttr, standardAttr);
							this.allFolders.Add(referenceNumber, currentFolder);
							currentFolder.SetParent(parentFolder); // Link it to the parent							
							return true;
						}
					}
					//else
					//{
					// System folder
					//}
				}
				else
				{
					// No name attribute found
				}
			}
			return false; // No folder added
		}


		/// <summary>
		/// Add the filesize to a folder
		/// And add the filesize to all parent folders
		/// Note: Will loop!
		/// </summary>
		/// <param name="allFolders"></param>
		/// <param name="parentId"></param>
		/// <param name="dataSize"></param>
		/// <returns></returns>
		public void AddFile(ulong fileReferenceNumber, ulong folderReferenceNumber, ulong fileSize, string fileName)
		{
			Folder currentFolder = this[folderReferenceNumber];
			if (currentFolder == null)
			{
				// Folder does not yet exist, create it
				// But we do not yet know the parent. Later when this parent is
				// encountered, its parent will be updated with this info
				currentFolder = new Folder(folderReferenceNumber);
				this.allFolders.Add(folderReferenceNumber, currentFolder);
			}

			// Add this file to the folder
			// And increase the size of the parents and parents-parents etc. (if available)
			currentFolder.AddFile(fileReferenceNumber, fileSize, fileName);
		}


		/// <summary>
		/// Build the tree
		/// </summary>
		/// <param name="allFolders"></param>
		public void FinalizeTree(NTFSReader parser)
		{
			// Loop through all folders, add the folders to their parents
			this.rootFolder = null;
			foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
			{
				if (kvp.Value.ParentFolder == null)
				{
					// There is only one where is is applicable, the root:
					this.rootFolder = kvp.Value;
					this.rootFolder.Name = parser.Drive;
					break;
				}
			}

			// Store pointer to reader in all folders (might take up a considerable amount of space...)
			foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
				kvp.Value.Parser = parser;
		}


		/// <summary>
		/// Compare the filesize
		/// </summary>
		/// <param name="a"></param>
		/// <param name="b"></param>
		/// <returns></returns>
		static int FileSizeCompare(KeyValuePair<File, ulong> a, KeyValuePair<File, ulong> b)
		{
			return -1 * a.Value.CompareTo(b.Value);
		}


		/// <summary>
		/// Find the largest X files
		/// </summary>
		/// <param name="allFolders"></param>
		public List<KeyValuePair<ulong, File>> FindLargest(int largestX = 100)
		{
			// Loop through all folders, 
			List<KeyValuePair<File, ulong>> allFiles = new List<KeyValuePair<File, ulong>>(); // < File, size >
			foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
			{
				Folder currentFolder = kvp.Value;
				if (currentFolder.Files != null)
				{
					foreach(KeyValuePair<ulong, File> fd in currentFolder.Files)
					{
						allFiles.Add(new KeyValuePair<File, ulong>(fd.Value, fd.Value.Size));
					}
				}
			}
			allFiles.Sort(FolderCollection.FileSizeCompare);

			// Now convert the a <reference, file> list
			return allFiles.Take(largestX).Select( (a, b) => new KeyValuePair<ulong, File>(a.Key.ReferenceNumber, a.Key )).ToList();
		}


		/// <summary>
		/// Find all files that match a searchmask (use wildcards ? and *)
		/// </summary>
		/// <param name="allFolders"></param>
		public List<File> FindFiles(string searchMask)
		{
			// Compiled regex 700ms => 500ms
			Regex regex = new Regex("^" + Regex.Escape(searchMask).Replace("\\*", ".*").Replace("\\?", ".") + "$", RegexOptions.IgnoreCase | RegexOptions.CultureInvariant | RegexOptions.Compiled);

			// Loop through all folders (+- 700 ms)
			Stopwatch sw = Stopwatch.StartNew();
			List<File> allFiles = new List<File>();
			foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
			{
				Folder currentFolder = kvp.Value;
				if (currentFolder.Files != null)
				{
					allFiles.AddRange(currentFolder.Files.Select(a => a.Value));
				}
			}

			Debug.WriteLine("Create list in {0} ms", sw.ElapsedMilliseconds);

			return allFiles.Where(a => regex.IsMatch(a.Name)).ToList();


			//foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
			//{
			//	Folder currentFolder = kvp.Value;
			//	if (currentFolder.Files != null)
			//	{
			//		foreach (KeyValuePair<ulong, File> fd in currentFolder.Files.Where(a => regex.IsMatch(a.Value.Name) ))
			//		{
			//			allFiles.Add(fd.Value);
			//		}
			//	}
			//}


			//foreach (KeyValuePair<ulong, Folder> kvp in this.allFolders)
			//{
			//	Folder currentFolder = kvp.Value;
			//	if (currentFolder.Files != null)
			//	{
			//		allFiles.AddRange(currentFolder.Files.Where(a => regex.IsMatch(a.Value.Name)).Select(a => a.Value));
			//	}
			//}

			//// +- 700ms
			//return allFiles;
		}
	}
}
