﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace Myriadbits.NTFS
{
	class UnmanagedFileLoader
	{
		public const uint FILE_ATTRIBUTE_NORMAL = 0x80;
		public const short INVALID_HANDLE_VALUE = -1;
		public const uint GENERIC_READ = 0x80000000;
		public const uint GENERIC_WRITE = 0x40000000;
		public const uint CREATE_NEW = 1;
		public const uint CREATE_ALWAYS = 2;
		public const uint OPEN_EXISTING = 3;
		public const uint FILE_SHARE_READ = 1;
		public const uint FILE_SHARE_WRITE = 2;
		public const short INVALID_SET_FILE_POINTER = -1;
		public const uint FILE_BEGIN = 0;
		private SafeFileHandle handleValue = null;

		// Use interop to call the CreateFile function.
		// For more information about CreateFile,
		// see the unmanaged MSDN reference library.
		[DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
		static extern SafeFileHandle CreateFile(string lpFileName, uint dwDesiredAccess,
		  uint dwShareMode, IntPtr lpSecurityAttributes, uint dwCreationDisposition,
		  uint dwFlagsAndAttributes, IntPtr hTemplateFile);


		[DllImport("kernel32.dll", SetLastError = true)]
		static extern bool ReadFile(IntPtr hFile, [Out] byte[] lpBuffer, uint nNumberOfBytesToRead, out uint lpNumberOfBytesRead, IntPtr lpOverlapped);

		[DllImport("kernel32.dll")]
		static extern bool SetFilePointerEx(IntPtr hFile, long liDistanceToMove, IntPtr lpNewFilePointer, uint dwMoveMethod);


		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="Path"></param>
		public UnmanagedFileLoader(string Path)
		{
			Load(Path);
		}


		/// <summary>
		/// Load a new file
		/// </summary>
		/// <param name="Path"></param>
		public void Load(string Path)
		{
			if (Path == null || Path.Length == 0)
			{
				throw new ArgumentNullException("Path");
			}

			// Try to open the file.
			handleValue = CreateFile(Path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero,
				OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, IntPtr.Zero);

			// If the handle is invalid,
			// get the last Win32 error 
			// and throw a Win32Exception.
			if (handleValue.IsInvalid)
			{
				Marshal.ThrowExceptionForHR(Marshal.GetHRForLastWin32Error());
			}
		}


		public uint ReadFile(ref byte[] lpBuffer, int numberOfBytesToRead)
		{
			uint numberOfBytesRead = 0;
			ReadFile(this.handleValue.DangerousGetHandle(), lpBuffer, (uint) numberOfBytesToRead, out numberOfBytesRead, IntPtr.Zero);
			return numberOfBytesRead;
		}

		public byte[] ReadBytes(uint numberOfBytesToRead)
		{
			byte[] data = new byte[numberOfBytesToRead];
			uint numberOfBytesRead = 0;
			ReadFile(this.handleValue.DangerousGetHandle(), data, numberOfBytesToRead, out numberOfBytesRead, IntPtr.Zero);
			return data;
		}

		public void Seek(ulong absolutePosition)
		{
			SetFilePointerEx(this.handleValue.DangerousGetHandle(), (long) absolutePosition, IntPtr.Zero, FILE_BEGIN);
		}

		/// <summary>
		/// Return the handle
		/// </summary>
		public SafeFileHandle Handle
		{
			get
			{
				// If the handle is valid,
				// return it.
				if (!handleValue.IsInvalid)
				{
					return handleValue;
				}
				else
				{
					return null;
				}
			}
		}
	}
}
